import aiohttp
import logging
import requests

from abc import ABC, abstractmethod
from typing import Any, Callable

from ..exceptions import BentoAuthException

__all__ = ["BaseAuthMiddleware"]


class BaseAuthMiddleware(ABC):
    def __init__(
        self,
        bento_authz_service_url: str,
        drs_compat: bool = False,
        sr_compat: bool = False,
        beacon_meta_callback: Callable[[], dict] | None = None,
        debug_mode: bool = False,
        enabled: bool = True,
        logger: logging.Logger | None = None,
    ):
        self._debug: bool = debug_mode
        self._verify_ssl: bool = not debug_mode

        self._enabled: bool = enabled
        self._logger: logging.Logger | None = logger

        self._drs_compat: bool = drs_compat
        self._sr_compat: bool = sr_compat
        self._beacon_meta_callback: Callable[[], dict] | None = beacon_meta_callback

        self._bento_authz_service_url: str = bento_authz_service_url

    @property
    def enabled(self) -> bool:
        return self._enabled

    @abstractmethod
    def get_authz_header_value(self, request: Any) -> str | None:  # pragma: no cover
        pass

    @staticmethod
    def check_require_token(require_token: bool, token: str | None) -> None:
        if require_token:
            if token is None:
                raise BentoAuthException("No token provided")

    def mk_authz_url(self, path: str) -> str:
        return f"{self._bento_authz_service_url.rstrip('/')}{path}"

    def _extract_token_and_build_headers(
        self,
        request: Any,
        require_token: bool,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> dict[str, str]:
        if headers_getter:
            return headers_getter(request)

        tkn_header = self.get_authz_header_value(request)
        self.check_require_token(require_token, tkn_header)
        return {"Authorization": tkn_header} if tkn_header else {}

    def _log_error(self, message: str):
        if self._logger:
            self._logger.error(message)

    def _gen_exc_non_200_error_from_authz(self, code: int, content: bytes):
        self._log_error(f"Got non-200 response from authorization service: {code} {content}")
        # Generic error - don't leak errors from authz service!
        raise BentoAuthException("Error from authz service", status_code=500)

    def authz_post(
        self,
        request: Any,
        path: str,
        body: dict,
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> dict:
        res = requests.post(
            self.mk_authz_url(path),
            json=body,
            headers=self._extract_token_and_build_headers(request, require_token, headers_getter),
            verify=self._verify_ssl)

        if res.status_code != 200:  # Invalid authorization service response
            raise self._gen_exc_non_200_error_from_authz(res.status_code, res.content)

        return res.json()

    async def async_authz_post(
        self,
        request: Any,
        path: str,
        body: dict,
        require_token: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ) -> dict:
        async with aiohttp.ClientSession() as session:
            async with session.post(
                    self.mk_authz_url(path),
                    json=body,
                    headers=self._extract_token_and_build_headers(request, require_token, headers_getter),
                    ssl=(None if self._verify_ssl else False)) as res:

                if res.status != 200:  # Invalid authorization service response
                    raise self._gen_exc_non_200_error_from_authz(res.status, await res.content.read())

                return await res.json()

    @staticmethod
    @abstractmethod
    def mark_authz_done(request: Any):  # pragma: no cover
        pass

    def check_authz_evaluate(
        self,
        request: Any,
        permissions: frozenset[str],
        resource: dict,
        require_token: bool = True,
        set_authz_flag: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ):
        if not self.enabled:
            return

        res = self.authz_post(
            request,
            "/policy/evaluate",
            body={"requested_resource": resource, "required_permissions": list(permissions)},
            require_token=require_token,
            headers_getter=headers_getter,
        )

        if not res.get("result"):
            # We early-return with the flag set - we're returning Forbidden,
            # and we've determined authz, so we can just set the flag.
            raise BentoAuthException("Forbidden", status_code=403)  # Actually forbidden by authz service

        if set_authz_flag:
            self.mark_authz_done(request)

    async def async_check_authz_evaluate(
        self,
        request: Any,
        permissions: frozenset[str],
        resource: dict,
        require_token: bool = True,
        set_authz_flag: bool = False,
        headers_getter: Callable[[Any], dict[str, str]] | None = None,
    ):
        if not self.enabled:
            return

        res = await self.async_authz_post(
            request,
            "/policy/evaluate",
            body={"requested_resource": resource, "required_permissions": list(permissions)},
            require_token=require_token,
            headers_getter=headers_getter,
        )

        if not res.get("result"):
            # We early-return with the flag set - we're returning Forbidden,
            # and we've determined authz, so we can just set the flag.
            raise BentoAuthException("Forbidden", status_code=403)  # Actually forbidden by authz service

        if set_authz_flag:
            self.mark_authz_done(request)



import os
import werkzeug.utils

from typing import Dict, List, Optional, Union

__all__ = [
    "WORKFLOW_TYPE_FILE",
    "WORKFLOW_TYPE_FILE_ARRAY",
    "WORKFLOW_TYPE_STRING",
    "WORKFLOW_TYPE_STRING_ARRAY",
    "WORKFLOW_TYPE_ENUM",
    "WORKFLOW_TYPE_ENUM_ARRAY",
    "WORKFLOW_TYPES",
    "WORKFLOW_FILE_TYPES",

    "file_with_prefix",
    "formatted_output",
    "namespaced_input",
    "make_output_params",
    "find_common_prefix",

    "secure_filename",
    "workflow_exists",
    "get_workflow",
    "get_workflow_resource",
]


WORKFLOW_TYPE_ARRAY_SUFFIX = "[]"


def array_of_type(workflow_type: str) -> str:
    return f"{workflow_type}{WORKFLOW_TYPE_ARRAY_SUFFIX}"


WORKFLOW_TYPE_FILE = "file"
WORKFLOW_TYPE_FILE_ARRAY = array_of_type(WORKFLOW_TYPE_FILE)
WORKFLOW_TYPE_STRING = "string"
WORKFLOW_TYPE_STRING_ARRAY = array_of_type("string")
WORKFLOW_TYPE_ENUM = "enum"
WORKFLOW_TYPE_ENUM_ARRAY = array_of_type("enum")

WORKFLOW_TYPES = frozenset((
    WORKFLOW_TYPE_FILE,
    WORKFLOW_TYPE_FILE_ARRAY,
    WORKFLOW_TYPE_STRING,
    WORKFLOW_TYPE_STRING_ARRAY,
    WORKFLOW_TYPE_ENUM,
    WORKFLOW_TYPE_ENUM_ARRAY,
))

WORKFLOW_FILE_TYPES = frozenset((WORKFLOW_TYPE_FILE, WORKFLOW_TYPE_FILE_ARRAY))


def file_with_prefix(file_name: str, prefix: Optional[int]) -> str:
    return f"{prefix}_{file_name}" if prefix is not None else file_name


def _output_file_name(file_name, output_params) -> Optional[str]:
    try:
        return secure_filename(file_name.format(**output_params))
    except KeyError:
        return None


def _output_file_name_array(file_template, map_from_input, output_params) -> Optional[List[str]]:
    try:
        return [secure_filename(file_template.format(v)) for v in output_params[map_from_input]]
    except KeyError:
        return None


def _output_string(string, output_params) -> Optional[str]:
    try:
        return string.format(**output_params)
    except KeyError:
        return None


def _output_string_array(string_template, map_from_input, output_params) -> Optional[List[str]]:
    try:
        return [string_template.format(v) for v in output_params[map_from_input]]
    except KeyError:
        return None


def formatted_output(output: dict, output_params: dict) -> Optional[Union[str, List[str]]]:
    if output["type"] == WORKFLOW_TYPE_FILE:
        return _output_file_name(output["value"], output_params)
    elif output["type"] == WORKFLOW_TYPE_FILE_ARRAY:
        return _output_file_name_array(output["value"], output["map_from_input"], output_params)
    elif output["type"] in (WORKFLOW_TYPE_STRING, WORKFLOW_TYPE_ENUM):  # TODO: Check enum values?
        return _output_string(output["value"], output_params)
    elif output["type"] in (WORKFLOW_TYPE_STRING_ARRAY, WORKFLOW_TYPE_ENUM_ARRAY):
        return _output_string_array(output["value"], output["map_from_input"], output_params)
    else:
        raise NotImplementedError


def namespaced_input(workflow_name: str, input_id: str) -> str:
    return f"{workflow_name}.{input_id}"


def make_output_params(workflow_id: str, workflow_params: dict, workflow_inputs: list) \
        -> Dict[str, Union[str, List[str]]]:
    # TODO: This can raise KeyError on os.path.basename(workflow_params[ni]) if ni is incorrect (e.g. missing the
    #  namespaced prefix.) This should be explicitly documented and perhaps ParameterException or something should be
    #  introduced to force custom handling for this exception.

    output_params = {}

    for i, input_spec in enumerate(workflow_inputs):
        if input_spec["type"] not in WORKFLOW_TYPES:
            raise NotImplementedError

        ni = namespaced_input(workflow_id, input_spec["id"])

        # If the input is not required and nothing was specified for it, skip it
        if not input_spec.get("required", True) and workflow_params.get(ni) is None:
            continue

        if input_spec["type"] in WORKFLOW_FILE_TYPES:
            # TODO: DOCS: Just file name without path...
            # TODO: Separate params for full path / path without drop_box stuff?

            output_params[input_spec["id"]] = ([os.path.basename(f) for f in workflow_params[ni]]
                                               if input_spec["type"].endswith(WORKFLOW_TYPE_ARRAY_SUFFIX)
                                               else os.path.basename(workflow_params[ni]))

        else:  # Primitive type or array of primitive type
            output_params[input_spec["id"]] = workflow_params[ni]

    return output_params


def _get_file_paths_from_output(base_path, output, output_params, prefix=None):
    fo = formatted_output(output, output_params)
    return (os.path.join(base_path, file_with_prefix(f, prefix)) for f in (fo if isinstance(fo, list) else [fo])
            if f is not None)


def find_common_prefix(base_path: str, workflow_metadata: dict, output_params: dict) -> Optional[int]:
    # Increase the prefix until a suitable one has been found, if needed

    prefix = None

    while True:
        duplicate_exists = False

        for output in workflow_metadata["outputs"]:
            # It only makes sense to deal with file outputs TODO: IS THIS TRUE?
            if output["type"] not in WORKFLOW_FILE_TYPES:
                continue

            duplicate_exists = duplicate_exists or any(
                os.path.exists(file_path)
                for file_path in _get_file_paths_from_output(
                    base_path, output, output_params, prefix))

            if duplicate_exists:
                # We already know this prefix isn't good enough, so stop
                # checking files until the next go-around.
                break

        if not duplicate_exists:
            # We've found a good prefix, exit the loop and the function
            return prefix

        # Otherwise, go around again to test a better prefix
        prefix = (prefix or 0) + 1


def secure_filename(fn: str) -> str:
    return werkzeug.utils.secure_filename(fn)


def workflow_exists(workflow_id: str, wfs: Dict):
    for key in wfs:
        if workflow_id in wfs[key]:
            return True

    return False


def get_workflow(workflow_id: str, wfs: Dict):
    for key in wfs:
        if workflow_id in wfs[key]:
            return wfs[key][workflow_id]

    raise KeyError(f'Workflow ID {workflow_id} not found.')


def get_workflow_resource(workflow_id: str, wfs: Dict):
    return secure_filename(get_workflow(workflow_id, wfs)["file"])



#!/usr/bin/env python

import configparser
import os
import setuptools

with open("README.md", "r") as rf:
    long_description = rf.read()

config = configparser.ConfigParser()
config.read(os.path.join(os.path.dirname(os.path.realpath(__file__)), "bento_lib", "package.cfg"))

setuptools.setup(
    name=config["package"]["name"],
    version=config["package"]["version"],

    python_requires=">=3.10.0",
    install_requires=[
        "aiohttp>=3.8.4,<4",
        "jsonschema>=4.17.3,<5",
        "psycopg2-binary>=2.9.5,<3.0",
        "redis>=4.5.4,<5.0",
        "requests>=2.28.1,<3",
        "Werkzeug>=2.2.3,<3",
    ],
    extras_require={
        "flask": ["Flask>=2.2.5,<3"],
        "django": ["Django>=4.2.1,<5", "djangorestframework>=3.14.0,<3.15"],
        "fastapi": ["fastapi>=0.100,<0.104"],
    },

    author=config["package"]["authors"],
    author_email=config["package"]["author_emails"],

    description="A set of common utilities and helpers for Bento platform services.",
    long_description=long_description,
    long_description_content_type="text/markdown",

    packages=setuptools.find_namespace_packages(),
    include_package_data=True,

    url="https://github.com/bento-platform/bento_lib",
    license="LGPLv3",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent"
    ]
)
